/* File:	  Library.P
** Author(s): Ninghui Li
** Contact:   ninghui@cs.nyu.edu
** 
** Copyright (C)  2000, Ninghui Li
**
** XD1LP is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XD1LP is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XD1LP; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


%% Note that the cut(!) in library isn't effective yet.

%% Conjunctions and Disjunctions
rule(holds((PS1,PS2), BAtom, Len),
     (holds(PS1, BAtom, Len), holds(PS2, BAtom, Len))).
rule(holds((PS1;PS2), BAtom, Len),
     ((holds(PS1, BAtom, Len); holds(PS2, BAtom, Len)))).
rule(delegates((PS1,PS2), BAtom, Dep, Dele, Len),
     (delegates(PS1, BAtom, Dep, Dele, Len),
      delegates(PS2, BAtom, Dep, Dele, Len))).
rule(delegates((PS1;PS2), BAtom, Dep, Dele, Len),
     ((delegates(PS1, BAtom, Dep, Dele, Len);
      delegates(PS2, BAtom, Dep, Dele, Len)))).


%% Static unweighted threshold structure for holds
rule(holds(suth(I, _PrinList), _BAtom, _Len),  I =< 0).
rule(holds(suth(I, [Prin|PrinList]), BAtom, Len), 
     (I > 0, 
      holds(Prin, BAtom, Len),
      Iminus1 is I - 1,
      holds(suth(Iminus1, PrinList), BAtom, Len))).
rule(holds(suth(I, [_Prin|PrinList]), BAtom, Len),
     (I > 0,
      holds(suth(I, PrinList), BAtom, Len))).

%% Static unweighted threshold structure for delegates
rule(delegates(suth(I, _PrinList), _BAtom, _Dep, _Dele, _Len),  I =< 0).
rule(delegates(suth(I, [Prin|PrinList]), BAtom, Dep, Dele, Len), 
     (I > 0,
      delegates(Prin, BAtom, Dep, Dele, Len),
      Iminus1 is I - 1,
      delegates(suth(Iminus1, PrinList), BAtom, Dep, Dele, Len))).
rule(delegates(suth(I, [_Prin|PrinList]), BAtom, Dep, Dele, Len),
     (I > 0,
      delegates(suth(I, PrinList), BAtom, Dep, Dele, Len))).

%% Static weighted threshold structure for holds
rule(holds(swth(I, _PWList), _BAtom, _Len), I =< 0).
rule(holds(swth(I, [(Prin,Weight)|PWList]), BAtom, Len), 
     (I > 0,  
      holds(Prin, BAtom, Len),
      IminusW is I - Weight,
      holds(swth(IminusW, PWList), BAtom, Len))).
rule(holds(swth(I, [(_Prin,_Weight)|PWList]), BAtom, Len),
     (I > 0,
      holds(swth(I, PWList), BAtom, Len))).

%% Static weighted threshold structure for delegates
rule(delegates(swth(I, _PWList), _BAtom, _Dep, _Dele, _Len),  I =< 0).
rule(delegates(swth(I, [(Prin,Weight)|PWList]), BAtom, Dep, Dele, Len), 
     (I > 0,
      delegates(Prin, BAtom, Dep, Dele, Len),
      IminusW is I - Weight, 
      delegates(swth(IminusW, PWList), BAtom, Dep, Dele, Len))).
rule(delegates(swth(I, [(_Prin,_Weight)|PWList]), BAtom, Dep, Dele, Len),
     (I > 0,
      delegates(swth(I, PWList), BAtom, Dep, Dele, Len))).

%% Dynamic unweighted threshold structure for holds
rule(holds(duth(I, _J, dholds(_DP, _DIssuer, _DBAtom)), _BAtom, _Len), 
     I =< 0).
rule(holds(duth(I, J, dholds(DP, DIssuer, DBAtom)), BAtom, Len), 
     (I>0,       %% Make a fresh copy
      copy_term(dholds(DP, DIssuer, DBAtom), dholds(NDP, NDIssuer, NDBAtom)),
      ith_issuer(J, NDP),              %% DP is the jth issuer
      holds(NDIssuer, NDBAtom, '*'),   %% DP satisfies the condition
      holds(NDP, BAtom, Len),
      Iminus1 is I - 1, 
      Jplus1 is J + 1,
      holds(duth(Iminus1, Jplus1, dholds(DP, DIssuer, DBAtom)), BAtom, Len))).
rule(holds(duth(I, J, dholds(DP, DIssuer, DBAtom)), BAtom, Len),
     (I>0,
      Jplus1 is J + 1,
      ith_issuer(Jplus1, _AnyIssuer),           %% there is a (j+1)th issuer
      holds(duth(I, Jplus1, dholds(DP, DIssuer, DBAtom)), BAtom, Len))).

%% Dynamic unweighted threshold structure for delegates
rule(delegates(duth(I,_J,dholds(_DP,_DIssuer,_DBAtom)),_BAtom,_Dep,_Dele,_Len),
     I =< 0).
rule(delegates(duth(I,J,dholds(DP,DIssuer,DBAtom)), BAtom,Dep,Dele,Len), 
     (I > 0,   %% Make a fresh copy
      copy_term(dholds(DP, DIssuer, DBAtom), dholds(NDP, NDIssuer, NDBAtom)),
      ith_issuer(J, NDP),                %% DP is the jth issuer
      holds(NDIssuer, NDBAtom, '*'),   %% DP satisfies the condition
      delegates(NDP, BAtom, Dep, Dele, Len),
      Iminus1 is I - 1, 
      Jplus1 is J + 1,
      delegates(duth(Iminus1, Jplus1, dholds(DP,DIssuer,DBAtom)), 
              BAtom,Dep,Dele, Len))).
rule(delegates(duth(I,J,dholds(DP,DIssuer,DBAtom)), BAtom,Dep,Dele,Len), 
     (I > 0,
      Jplus1 is J + 1,
      ith_issuer(Jplus1, _AnyIssuer),   
      delegates(duth(I,Jplus1,dholds(DP,DIssuer,DBAtom)),BAtom,Dep,Dele,Len))).


%% Dynamic weighted threshold structure for holds
rule(holds(dwth(I, _J, dholds(_DP, _DW, _DIssuer, _DBAtom)), _BAtom, _Len), 
     I =< 0).
rule(holds(dwth(I, J, dholds(DP, DW, DIssuer, DBAtom)), BAtom, Len), 
     (I>0,       %% Make a fresh copy
      copy_term(dholds(DP,DW,DIssuer,DBAtom), dholds(NDP,NDW,NDIssuer,NDBAtom)),
      ith_issuer(J, NDP),              %% DP is the jth issuer
      holds(NDIssuer, NDBAtom, '*'),   %% DP satisfies the condition
      integer(NDW),
      holds(NDP, BAtom, Len),
      IminusW is I - NDW, 
      Jplus1 is J + 1,
      holds(dwth(IminusW, Jplus1, dholds(DP, DW, DIssuer, DBAtom)), BAtom, Len))).
rule(holds(dwth(I, J, dholds(DP, DW, DIssuer, DBAtom)), BAtom, Len),
     (I>0,
      Jplus1 is J + 1,
      ith_issuer(Jplus1, _AnyIssuer),           %% there is a (j+1)th issuer
      holds(duth(I, Jplus1, dholds(DP, DW, DIssuer, DBAtom)), BAtom, Len))).

%% Dynamic weighted threshold structure for delegates
rule(delegates(dwth(I,_J, dholds(_DP,_DW,_DIssuer,_DBAtom)),_BAtom,_Dep,_Dele,_Len),
     I =< 0).
rule(delegates(dwth(I, J, dholds(DP, DW, DIssuer, DBAtom)), BAtom, Dep, Dele, Len), 
     (I > 0,    %% Make a fresh copy
      copy_term(dholds(DP,DW,DIssuer,DBAtom), dholds(NDP,NDW,DIssuer,NDBAtom)),
      ith_issuer(J, NDP),             %% DP is the jth issuer
      holds(DIssuer, NDBAtom, '*'),   %% DP satisfies the condition
      integer(NDW),
      delegates(NDP, BAtom, Dep, Dele, Len), 
      IminusW is I - NDW, 
      Jplus1 is J + 1, 
      delegates(dwth(IminusW, Jplus1, dholds(DP, DIssuer, DBAtom)), BAtom, Dep, Dele, Len))).
rule(delegates(dwth(I, J, dholds(DP, DW, DIssuer, DBAtom)), BAtom, Dep, Dele, Len), 
     (I > 0,
      Jplus1 is J + 1,
      ith_issuer(Jplus1, _AnyIssuer), 
      delegates(dwth(I, Jplus1, dholds(DP, DW, DIssuer, DBAtom)), BAtom, Dep, Dele, Len))).
