/* File:	  Main.P
** Author(s): Ninghui Li
** Contact:   ninghui@cs.nyu.edu
** 
** Copyright (C)  2000, Ninghui Li
**
** XD1LP is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XD1LP is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XD1LP; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


:- compiler_options([xpp_on]).

:- import append/3, ith/3, member/2 from basics.
:- import ensure_dyn_loaded/1 from consult.

:- import string_to_tokens/2 from 'Tokenizer'.
:- import compile_rule/4 from 'Compiler'.

:- assert(library_directory('c:\work\dl')).
:- ensure_dyn_loaded('C:\work\dl\library.P').
:- initialize_db.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  The following are predicates for testing purpose 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

load :- consult('c:\WORK\DL\Main').

deb :- abolish_all_tables, trace.

test1 :- clear_db, add_new_module(alice_credit, file, 'c:\work\DL\examples\test1.d1lp').

query10(S) :- tsetof([BA,D,P], prove(delegates('Alice', BA, D, P, '*')), S).
query11 :- prove(delegates('Alice', purchase(a,100), 1, 'Carl', '*')).
query16 :- prove(delegates('Alice', purchase(a,1000), 1, 'Carl', '*')).
query12 :- prove(holds(keyBob, goodCredit('Carl'), '*')).
query13 :- prove(delegates('Bob', keyBob, goodCredit(_X), '*', 0)).
query14 :- prove(holds('Alice', goodCredit('Carl'), '*')).
query15 :- prove(holds(cardC, accountGood('Carl'), '*')).

test2 :- clear_db, add_new_module(hospital, file, 'c:\work\DL\examples\test2.d1lp').
  %% This one doesn't work yet.  Don't know why.
query20(S) :- tsetof(authorized(A,R), prove(holds('HM', authorized(A,R), '*')), S). 
query21 :- prove(holds('HM', authorized('Alice', read(medRec('Peter'))), '*')).
query22 :- prove(holds('HM', inRole('HA', hospital), '*')).
query23 :- prove(holds('HM', inRole('Alice', physician('Peter')), '*')).
query24 :- prove(holds('HM', authorized('Bob', read(medRec('Peter'))), '*')).
query25 :- prove(holds('HA', inRole('Alice', physician('Peter')), 1)).
query26 :- prove(holds('HA', inRole('Alice', physician('Peter')), '*')).
query27 :- prove(holds('HM', inRole('HA', hospital), 2)).


test3 :- clear_db, add_new_module(mpki, file, 'c:\work\DL\examples\test3.d1lp').
query30(S) :- tsetof(BA, prove(holds('Alice', BA, '*')), S).
query39(S) :- tsetof([BA,D,P], prove(delegates('Alice', BA, D, P, '*')), S).
query31 :- prove(holds('Alice', isSiteKey('MKey', 'MSite'), '*')).
query32 :- prove(holds('Alice', isSiteKey('LKey', 'LSite'), '*')).
query33 :- prove(holds('Bob', belongsTo('MSite', orga), '*')).
query34 :- prove(holds('Bob', isSiteKey('MKey', 'MSite'), '*')).

test4 :- clear_db, add_new_module(control, file, 'c:\work\DL\examples\test4.d1lp').
query40(S) :- tsetof(BA, prove(holds('Alice', BA, '*')), S).
query49(S) :- tsetof([BA,D,P], prove(delegates('Alice', BA, D, P, '*')), S).
query48(S) :- tsetof([BA,D,P], prove(delegates(tmpKey, BA, D, P, '*')), S).
query41 :- prove(delegates('Alice', access, 1, 'David', 2)).
query42 :- prove(delegates(tmpKey, access, 1, 'David', 1)).

test5 :- clear_db, add_new_module(long, file, 'c:\work\DL\examples\test11.d1lp').
query50(S) :- tsetof(BA, prove(holds('Alice', BA, '*')), S).
query51 :- prove(holds('Alice', isSiteKey('Key9', 'Site9'), '*')).

time(Goal, T) :- cputime(T1), call(Goal), cputime(T2), T is T2 - T1.

update :- compile_files('C:\WORK\DL\', ['Tokenizer', 'Compiler', 'Library']).
  %% The predicate to load all components
compile_files(_Dir, []).
compile_files(Dir, [File | FileList]) :- 
    atom_codes(Dir, DirCodes),
    atom_codes(File, FileCodes),
    append(DirCodes, FileCodes, PathCodes),
    atom_codes(Path, PathCodes),
    compile(Path),
    compile_files(Dir, FileList).

initialize_db :- clear_db.

clear_db :-
    abolish_all_tables,
    retractall(history(_)),
    retractall(max_depth(_)),
    retractall(rule(_Head, _Body, _ModName)),
    retractall(srule(_Head, _Body, _ModName)),
    retractall(sfact(_Head, _ModName)),
    retractall(ith_issuer(_I, _Prin)),
    retractall(new_atom_counter(_N)),
    assertz(new_atom_counter(0)),
    assert(history([])),
    assert(max_depth(1)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The following is the inference engine, the main predicate is prove/1
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- table(prove/1).
:- table(minus_one/2).
:- table(plus_one/2).
:- table(special_minus_one/2).
:- table(less_equal/2).
:- table(ith_issuer/2).
:- table(is_prin_or_var/1).
:- table(special_minus_one/2).
:- table(special_minus_one_min/3).
:- table(special_less_tran/2).


prove((A,B)) :- !, prove(A), prove(B).
prove((A;B)) :- !, (prove(A); prove(B)).
prove(true) :- !.
prove(!) :- !, true.
prove((X is Y)) :- !, call(X is Y).
prove((X=<Y)) :- !, X =< Y.
prove((X>Y)) :- !, X > Y.
prove(copy_term(X, Y)) :- !, copy_term(X,Y).

prove(special_minus_one(A, B)) :- !, special_minus_one(A, B).
prove(special_minus_one_min(A,B,C)) :- !, special_minus_one_min(A,B,C).
prove(special_less_tran(A, B)) :- !, special_less_tran(A,B).
prove(less_equal(A, B)) :- !, less_equal(A,B).
prove(is_prin_or_var(P)) :- !, is_prin_or_var(P).
prove(ith_issuer(I, P)) :- !, ith_issuer(I, P).
prove(minus_one(X,Y)) :- !, minus_one(X, Y).
prove(plus_one(X,Y)) :- !, plus_one(X, Y).

  %% For system rules
prove(holds('System', B, _L)) :- 
    !, call(B).

  %% For efficency reasons, we divide rules into four groups, 
  %% rule/2 for system wide rules,  
  %% rule/3 for generated chaining rules with variable head lengths
  %% sfact/2 for generated facts, they always have fixed head lengths
  %% srule/3 for generated rules that have fixed head lengths

  %% Assume that L is always in [1..*] for holds.
  %% Assume that sfact and srule always have head length 1
#ifdef BEDUG_PROVER
prove(holds(I,B,L)) :- 
    less_equal(L,0), exception('Wrong parameter', holds(I,B,L)), fail.  
    %% testing only
#endif

prove(holds(I,B,L)) :- 
    !, ( sfact(holds(I,B,1), _ModName) ;
         (srule(holds(I,B,1), Body, _ModName), prove(Body)) ;
         (rule(holds(I,B,L), Body, _ModName), prove(Body)) ; 
         (rule(holds(I,B,L), Body), prove(Body)) ).

  %% Assume that L is always in [0..*] for delegates.
  %% Assume that sfact and srule always have head length 0 or 1
  %% Perhaps, we can deal with self-delegation in the prove level, rather than
  %% the meta rule level.
prove(delegates(I,B,D,E,L)) :- 
    is_prin_or_var(I), 
    ( I=D ; 
      (sfact(delegates(I,B,D1,E,L1), _ModName), less_equal(D,D1), less_equal(L1,L)) ;
      (srule(delegates(I,B,D,E,L1), Body, _ModName), less_equal(L1,L), prove(Body)) ;
      (rule(delegates(I,B,D,E,L), Body, _ModName), prove(Body)) ).

prove(delegates(I,B,D,E,L)) :- 
    !, (rule(delegates(I,B,D,E,L), Body), prove(Body)). 

prove(A) :-  exception('undefined predicate', A).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  The following is misc. utility predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  % expand_lines(+LineList, String)
  %   convert a list of char lists to one char list.
expand_lines([], []) :- !.
expand_lines([[] | Lines], String) :- 
    !,
    expand_lines(Lines, String).
expand_lines([[C | RestOfLine] | Lines], [C|String]) :- 
    !,
    expand_lines([RestOfLine | Lines], String).

list_end([E], E) :- !.
list_end([_H|L], E) :- list_end(L, E).

/* Don't move the following three predicates to Utils, they need to access
   max_dep */
minus_one('*', D) :- !, max_depth(D).
minus_one(N, NMinus1) :- integer(N), N>0, NMinus1 is N - 1.

plus_one(N, '*') :- integer(N), max_depth(D), N>=D, !.
plus_one(N, NPlus1) :- integer(N), !, NPlus1 is N + 1.

less_equal(_A, '*') :- !.
less_equal('*', _N) :- !, fail. 
less_equal(M, N) :- M =< N.

 %% The first parm. of special_minus_one is required to be ground
special_minus_one('*', '*') :- !.
special_minus_one(M, Mminus1) :- integer(M), !, Mminus1 is M-1.
#ifdef DEBUG_PROVER
special_minus_one(A, B) :- 
    exception('Illegal parameters', special_minus_one(A,B)), fail.
#endif

 %% The second parm. of special_less_tran is always ground
special_less_tran(_A, '*') :- !.
special_less_tran('*', _N) :- !, fail. 
special_less_tran(M, N) :- integer(M), !, M<N.
special_less_tran(M, N) :- var(M), !, M is N - 1.
#ifdef DEBUG_PROVER
special_less_tran(A,B) :- 
    exception('Illegal parameters', special_less_tran(A,B)), fail.
#endif

 %% Compute min(Len - 1 (special minus), Dep)
 %% Assume the first two parms are ground.
special_minus_one_min(N, '*', NN) :- !, special_minus_one(N, NN).
special_minus_one_min('*', L, L) :- !, integer(L).
special_minus_one_min(N, L, NN) :- 
#ifdef DEBUG_PROVER
    integer(N), integer(L), !, 
#endif
    (N>L -> NN = L; NN is N - 1).
#ifdef DEBUG_PROVER
special_minus_one_min(N, L, NN) :- 
    exception('Illegal parameters', special_minus_one_min(N,L,NN)), fail.
#endif

is_prin_or_var(Prin) :- is_principal(Prin); var(Prin).
is_principal(Prin) :- atom(Prin).

integer_max(M, N, M)  :- integer(M), (integer(N) -> (M>=N, !); !).
integer_max(_M, N, N) :- integer(N), !.
#ifdef DEBUG_PROVER
integer_max(M, N, X)  :- 
    exception('Illegal parameters', integer_max(M,N,X)), fail.
#endif

exception(S, T) :- display('Exception: '), display(S), display(': '), display(T), nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The following deal with D1LP rules and queries in a module way.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 %% Add a new module to the D1LP inferencing system
add_new_module(ModName, SourceType, Source) :-
    retractall(d1lp_module(ModName, _MD, _IL)),
    retractall(rule(_Head, _Body, ModName)),
    retractall(srule(_Head, _Body, ModName)),
    retractall(sfact(_Head, ModName)),
    retractall(max_depth(_)),
    retractall(ith_issuer(_, _)),
    assert(d1lp_module(ModName, 1, [])),
    add_rules(ModName, SourceType, Source),
    set_status(ModName).

set_status(ModName) :-
    d1lp_module(ModName, MD, IL),
    assert(max_depth(MD)),
    assert_issuer_list(1, IL).

assert_issuer_list(_I, []) :- !.
assert_issuer_list(I, [H|T]) :- 
    !, assert(ith_issuer(I, H)), Iplus1 is I + 1, assert_issuer_list(Iplus1, T).

 %% this is called after each rule is compiled and analyzed
update_module_info(ModName, NMD, NIL) :-
    d1lp_module(ModName, OMD, OIL),
    integer_max(NMD, OMD, MD),
    my_merge(NIL, OIL, IL, Changed),
    (MD>OMD -> set_module_info(ModName, MD, IL) ;
     (Changed = yes -> set_module_info(ModName, MD, IL) ;
            true)).

my_merge([], L1, L1, no) :- !.
my_merge([H|T], L1, L3, Changed) :- 
    member(H, L1), !, my_merge(T, L1, L3, Changed).
my_merge([H|T], L1, L3, yes) :- 
    L3 = [H | L4], my_merge(T, L1, L4, _Changed).


set_module_info(ModName, NMD, NIL) :-
    retractall(d1lp_module(ModName, _MD, _IL)),
    assert(d1lp_module(ModName, NMD, NIL)).

 %% read_and_compile_rules(+type, +inputSource, -OLPRuleSetList)
 %% reads D1LP rules and them compile them into OLP rulesets
 %% (type=file, inputSource=fileName) or (type=atom, inputSource=theAtom)
add_rules(ModName, file, FileName) :-
    file_open(FileName, 'r', FileHandle),
    add_rules_from_opened_file(ModName, FileHandle),
    file_close(FileHandle).
add_rules(ModName, atom, AtomString) :-
    file_open(AtomString, 'sr', FileHandle),
    add_rules_from_opened_file(ModName, FileHandle),
    file_close(FileHandle).

add_rules_from_opened_file(ModName, FileHandle) :-
    repeat,
    read_one_rule(FileHandle, TokenListList, EndFlag),
    process_one_rule(ModName, TokenListList, EndFlag),
    EndFlag = 1.

process_one_rule(_ModName, _TokenListList, 1) :- !.
process_one_rule(ModName, TokenListList, _EndFlag) :- 
    expand_lines(TokenListList, Tokens),
    parse_rule(Tokens, D1LPRule),
    compile_rule(D1LPRule, OLPRuleSet, NMD, NIL),
    update_module_info(ModName, NMD, NIL),
    assert_module_rules(ModName, OLPRuleSet).

 %% note that when IsFullLine=1, it may be the case that the line is too long.
 %% currently, we just ignore the case
read_one_rule(FileHandle, [TokenList | TokenListList], EndFlag) :-
    file_read_line_list(FileHandle, Line, IsFullLine), 
    !,
    string_to_tokens(Line, TokenList),
    (IsFullLine=0 -> (TokenListList=[], EndFlag=1) ;       %% Just read the last line
                     (list_end(TokenList, '.') -> (TokenListList=[], EndFlag=0) ;
                         read_one_rule(FileHandle, TokenListList, EndFlag))).
  %% exceution reaches here when file_read_line_list fails
read_one_rule(_FileHandle, [], 1).

assert_module_rules(_ModName, []) :- !.
assert_module_rules(ModName, [RS | RSL]) :-
    !, assert_module_rules(ModName, RS),
       assert_module_rules(ModName, RSL).
assert_module_rules(ModName, srule(Head, true)) :-
    !, assert(sfact(Head, ModName)).
assert_module_rules(ModName, srule(Head, Body)) :-
    !, assert(srule(Head, Body, ModName)).
assert_module_rules(ModName, rule(Head, Body)) :-
    !, assert(rule(Head, Body, ModName)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  The following is the grammar for parsing D1LP programs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 %% Grammar for parsing the input 
parse_rule(Tokens, D1LPRule)   :- phrase(dlrule(D1LPRule), Tokens).
parse_query(Tokens, D1LPQuery) :- phrase(dlrule(D1LPQuery), Tokens).

  %% <dlrule> ::= <head> <body> ['.']
dlrule((Stmt :- RHS)) -->
#ifdef DEBUG_PARSER
    {trace},
#endif
    headstmt(Stmt, [], VT1),
    body(RHS, VT1, _),
    ['.'].

  %% <query> ::= <body> ['?']
dlquery(RHS) -->
    body(RHS, [], _),
    ['?'].

  %% <head> ::= <holdsstmt> | <delegatesstmt>
headstmt(Stmt, VTi, VTo) -->
    prinorvar(Prin, VTi, VT1),    %% parse a principal or a var
    ( holdsstmtrest(Stmt, Prin, VT1, VTo); 
      delegatesstmtrest(Stmt, Prin, VT1, VTo);
      speaksforstmtrest(Stmt, Prin, VT1, VTo) ).

  %% <body> ::= <bodystmt> (',' <bodystmt>)* 
body((Stmt, Stmts), VTi, VTo) -->
    imply, 
    bodystmt(Stmt, VTi, VT1), 
    bodyandstmts(Stmts, VT1, VTo).
body(true, VT, VT) --> 
    []. 

bodyandstmts((Stmt, Stmts), VTi, VTo) -->
    [','],
    bodystmt(Stmt, VTi, VT1),
    bodyandstmts(Stmts, VT1, VTo).
bodyandstmts(true, _VTi, _VTo) -->
    [].

  %% <bodystmt> ::= <holdsstmt> | <bodydelegatesstmt>
bodystmt(Stmt, VTi, VTo) -->
    psorvar(Issuer, VTi, VT1),
    ( holdsstmtrest(Stmt, Issuer, VT1, VTo);
      bodydelegatesstmtrest(Stmt, Issuer, VT1, VTo) ).

  %% <holdsstmt> ::= <principal> says PredName <params> 
holdsstmtrest(dl_holds(Issuer, BAtom), Issuer, VTi, VTo) -->
    [says],
    baseatom(BAtom, VTi, VTo).

  %% <delegatesstmt> ::= <principal> delegates PredName <params> ^ <depth> 
  %%                                    to <prinstruct>
delegatesstmtrest(dl_delegates(Issuer, BAtom, K, Delegatee), 
                  Issuer, VTi, VTo) -->
    [delegates],
    baseatom(BAtom, VTi, VT2),
    ['^'],
    depth(K),
    [to],
    psorvar(Delegatee, VT2, VTo).

speaksforstmtrest(dl_speaksfor(Delegatee, BAtom, Issuer),
                  Delegatee, VTi, VTo) -->
    [speaks_for],
    prinorvar(Issuer, VTi, VT1),
    [on],
    baseatom(BAtom, VT1, VTo).

  %% <bodydelegatesstmt> ::= <principal> delegates PredName <params> 
  %%                                   ^ <depth>  to <principal>
bodydelegatesstmtrest(dl_delegates(Issuer, BAtom, K, Delegatee), 
                      Issuer, VTi, VTo) -->
    [delegates],
    baseatom(BAtom, VTi, VT2),
    ['^'],
    depth(K),
    pcorvar(Delegatee, VT2, VTo).

baseatom(BAtom, VTi, VTo) -->
    [PredName],
    params(ParamList, VTi, VTo),
    {BAtom =.. [PredName | ParamList]}. 

params([Param|Paraml], VTi, VTo) -->
    ['('], {!},
    term(Param, VTi, VT1),
    paraml(Paraml, VT1, VTo).
params([], VT, VT) --> 
    [].

paraml([], VT, VT) --> [')'], {!}.
paraml([Param|Paraml], VTi, VTo) -->
    [','], {!},
    term(Param, VTi, VT1),
    paraml(Paraml, VT1, VTo).

term(X0, VTi, VTo) --> 
    var(X0, VTi, VTo), {!}.
term(Term, VTi, VTo) --> 
    [FuncName],
    params(Params, VTi, VTo),
    {Term =.. [FuncName | Params]}.

depth('*') --> ['*'], {!}.
depth(K) -->   int(K).

var(X0, VTi, VTo) -->
    [var(X)],
    { !,
      lookup(VTi,VTo,X,X0) }.	/* Handle variables. */

imply --> [if].
imply --> [':', '-'].
imply --> ['<', '-'].

int(K) --> [K], {integer(K), K>0}.

/*
 * Begin of Principal Structure Grammar
 */

prinorvar(P, VTi, VTo) --> var(P, VTi, VTo), {!}.
prinorvar(P, VT, VT)   --> principal(P).

psorvar(PS, VTi, VTo) --> var(PS, VTi, VTo), {!}.
psorvar(PS, VTi, VTo) --> prinstruct(PS, VTi, VTo).

pcorvar(PC, VTi, VTo) --> var(PC, VTi, VTo), {!}.
pcorvar(PC, VT, VT) --> ['('], princonj(PC, VT, VT), [')'].

  %% <prinstruct> ::= '(' <prinexpr> ')' | <thresholdstruct> | <principal>
prinstruct(PS, VTi, VTo) -->  ['('], {!}, prinexpr(PS, VTi, VTo), [')'].
prinstruct(PS, VTi, VTo) -->  
    [threshold], {!}, ['('], int(K), [','],
    thresholdstructrest(PS, K, VTi, VTo).
prinstruct(PS, VT, VT) -->  principal(PS).

  %% <prinexpr> ::= <prinstruct> | <prinstruct> <princonjunct> 
  %%              | <prinstruct> <prindisjunct>
prinexpr(PS, VTi, VTo) -->  
    prinstruct(PS, VTi, VTo). 
prinexpr((PS1,PS2), VTi, VTo) -->  
    prinstruct(PS1, VTi, VT1), 
    [','],
    prinexpr(PS2, VT1, VTo). 
prinexpr((PS1;PS2), VTi, VTo) -->
    prinstruct(PS1, VTi, VT1), 
    [';'],
    prinexpr(PS2, VT1, VTo). 

princonj(P, VT, VT) -->
    principal(P), {!}.
princonj((PS1,PS2), VTi, VTo) -->
    princonj(PS1, VTi, VT1),
    princonj(PS2, VT1, VTo).

thresholdstructrest(suth(K, L), K, VT, VT) -->
    prinlist(_N, L),
    [')'].
thresholdstructrest(swth(K, L), K, VT, VT) -->
    pwpairlist(_Ws, L),
    [')'].
thresholdstructrest(duth(K,1,dholds(DP, Prin, BAtom)), K, VTi, VTo) -->
    var(DP, VTi, VT1),  %% No checking whether prin var DP appears
    [','],              %% in BAtom
    principal(Prin), 
    [says], 
    baseatom(BAtom, VT1, VTo),
    [')']. 

thresholdstructrest(dwth(K,1,dholds(DP, DW, Prin, BAtom)), K, VTi, VTo) -->
    var(DP, VTi, VT1),
    [','],
    var(DW, VT1, VT2),
    [','],
    principal(Prin), 
    [says], 
    baseatom(BAtom, VT2, VTo),
    [')']. 

  %% <prinlist> ::= '[' <principal> (',' <principal>)* ']'
prinlist(Nplus1, [P|L]) -->
    ['['],
    principal(P),
    prinlistrest(N, L),
    {Nplus1 is N + 1}.
prinlistrest(Nplus1, [P|L]) -->
    [','], {!},
    principal(P),
    prinlistrest(N, L),
    {Nplus1 is N + 1}.
prinlistrest(0, []) --> 
    [']'].


  %% <pwpairlist> ::= '[' <pwpair> (',' <pwpair>)* ']'
pwpairlist(Wsum, [[Prin, Weight]|L]) -->
    ['['],
    pwpair(Prin, Weight),
    pwpairlistrest(Wrest, L),
    {Wsum is Wrest + Weight}.
pwpairlistrest(Wsum, [[Prin|Weight]|L]) -->
    [','], {!},
    pwpair(Prin, Weight),
    pwpairlistrest(Wrest, L),
    {Wsum is Wrest + Weight}.
pwpairlistrest(0, []) -->  [']'].

pwpair(Prin, Weight) --> ['('], principal(Prin), [','], int(Weight), [')'].


principal(P) --> [P], {atom(P)}.

lookup([],[(Name,Value)], Name,Value) :- !.
lookup([(Name,Value)|T],[(Name,Value)|T], Name, Value) :- !.
lookup([X|T1], [X|T2], Name, Value) :- lookup(T1,T2,Name,Value).
/*
 * End of Principal Structure Grammar
 */

/* 
 * notes:
 * 1. DCG rules for parsing is also in this file, because it just won't
 *    work to put it into a module.
 * 2. Do not use dynamic in XSB, dynamic seems to make it static
 *
 */

