/* File:    TopLevelWindow.java
**            A Swing-based Prolog listener
** Author(s): Miguel Calejo
** Contact:   interprolog@servisoft.pt
** 
** Copyright (C)  1998, Servisoft II - Desenvolvimento de Software, Lda., Portugal
** 
** InterProlog is free software; you can redistribute it and/or modify it
** under the terms of the GNU Library General Public License as published
** by the Free Software Foundation; either version 2 of the License, or (at
** your option) any later version.
** 
** InterProlog is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with InterProlog; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id$
** 
*/
package pt.servisoft.interprolog.gui;

import pt.servisoft.interprolog.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.table.*;
import javax.swing.event.*;

public class TopLevelWindow extends JFrame implements PrologOutputListener{    
    PrologEngine engine = null;
    JTextArea prologOutput, prologInput; 
    JMenu historyMenu; 
    private static int topLevelCount = 0;
    static boolean debug=false;

    private Vector historyVector = new Vector();
    private int indexOfHistory = 0;

    TopLevelWindow(){
        super("XSB listener (Swing)");
        constructWindowContents();
        constructMenu();

        historyVector.add(""); 
       
        engine = new PrologEngine();
        engine.sendAndFlush("write('Welcome to an InterProlog/XSB top level!').\n");
        if (engine==null) dispose(); // no interface object permitted!
        else topLevelCount++;
        debug=engine.isDebug();
        engine.addPrologOutputListener(this); // so we get output and prompt "events"
                
        addWindowListener( new WindowAdapter() {
            public void windowClosing(WindowEvent e){
                dispose();
                engine.shutdown();
                topLevelCount--;
                if (topLevelCount <= 0) System.exit(0);
                    // should check whether editor windows are changed...
            }
            
            public void windowActivated(WindowEvent e){
                prologInput.requestFocus();
            }
        });
        prologInput.addKeyListener(new KeyAdapter(){
            public void keyPressed(KeyEvent e){
                int key = e.getKeyCode();
                switch (key) {
                case KeyEvent.VK_ENTER:
                    sendToProlog();
                    e.consume();
                    break;
                case KeyEvent.VK_UP:
                    if (indexOfHistory > 0) {
                        historyVector.set(indexOfHistory, prologInput.getText());
                        indexOfHistory--;
                        prologInput.setText((String)historyVector.elementAt(indexOfHistory));
                    }
                    e.consume();
                    break;
                case KeyEvent.VK_DOWN:
                    if (indexOfHistory < historyVector.size()-1) {
                        indexOfHistory++;
                        prologInput.setText((String)historyVector.elementAt(indexOfHistory));
                    }
                    e.consume();
                    break;
                }
            }
        });
        show();
        focusInput();    
    }
    
    public PrologEngine getEngine(){
        return engine;
    }
    
      // For some strange reason, when using the portuguese
      // keyboard, characters produced with the help of the
      // AltGr key don't show up in JTextComponent's; this
      // solves the problem.
    private static final JTextComponent.KeyBinding[] altGrBindings = {
      new JTextComponent.KeyBinding( KeyStroke.getKeyStroke('['), "insert-content" ),
      new JTextComponent.KeyBinding( KeyStroke.getKeyStroke(']'), "insert-content" ),
      new JTextComponent.KeyBinding( KeyStroke.getKeyStroke('{'), "insert-content" ),
      new JTextComponent.KeyBinding( KeyStroke.getKeyStroke('}'), "insert-content" ),
      new JTextComponent.KeyBinding( KeyStroke.getKeyStroke('@'), "insert-content" ),
      new JTextComponent.KeyBinding( KeyStroke.getKeyStroke(''), "insert-content" ),
      new JTextComponent.KeyBinding( KeyStroke.getKeyStroke(''), "insert-content" )
    };

    void constructWindowContents(){
        Font prologFont = new Font("Courier",Font.PLAIN,12);
        Container c = getContentPane();
        c.setLayout(new BorderLayout());

        prologOutput = new JTextArea(20,80); prologOutput.setFont(prologFont); 
        prologOutput.setEditable(false); 
        prologOutput.setToolTipText("Here's Prolog standard output");
        // prologOutput.setLineWrap(true);  Swing crashes with large ammounts of text...
        prologOutput.setDoubleBuffered(true); // Use Swing double screen buffer

        JScrollPane piscroller = new JScrollPane();
        prologInput = new JTextArea(2,80); 
        prologInput.setFont(prologFont); 
        prologInput.setLineWrap(true);
        prologInput.setToolTipText("Prolog input, sent when you press enter");
        piscroller.getViewport().add(prologInput); 

        // add those few precious key bindings...
        JTextComponent.loadKeymap(prologInput.getKeymap(), altGrBindings, prologInput.getActions());

        JScrollPane scroller = new JScrollPane();
        scroller.getViewport().add(prologOutput);
        JSplitPane j = new JSplitPane (JSplitPane.VERTICAL_SPLIT, scroller, prologInput);
        c.add("Center",j);
        setSize(800,400);
        j.setDividerLocation(300);
        //j.resetToPreferredSizes();
        validate();
    }
    
    void constructMenu(){
        JMenuBar mb; 
        mb = new JMenuBar(); setJMenuBar(mb);
        
        JMenu lookMenu = new JMenu("Looks"); 
        
        final UIManager.LookAndFeelInfo[] availableLooks = UIManager.getInstalledLookAndFeels();
        for (int looki=0; looki<availableLooks.length; looki++){
            JMenuItem temp = new JMenuItem(availableLooks[looki].getName());
            lookMenu.add(temp);
            final int hack = looki;
            temp.addActionListener(new ActionListener(){
                public void actionPerformed(ActionEvent e){
                    try{
                        UIManager.setLookAndFeel(availableLooks[hack].getClassName());
                        SwingUtilities.updateComponentTreeUI(getContentPane());
                        // should notify everybody...
                    } catch(Exception foo) {
                        System.out.println("You are not allowed to change to that look and feel!");
                    }
                }
            } );
        }
        JMenu toolMenu = new JMenu("Tools"); mb.add(toolMenu);
        mb.add(lookMenu);
                
        /*
        JMenuItem testSerialization = new JMenuItem("Test serialization to Prolog...");
        toolMenu.add(testSerialization);
        testSerialization.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e){
                try{                                        
                    FileOutputStream o = new FileOutputStream("outStream");
                    ObjectOutputStream out = new ObjectOutputStream(o);
                    //out.writeObject(prologInput);
                    //out.writeObject(list1); out.writeObject(list2); 
                    //out.writeObject(rp); 
                    //out.writeObject(mp); 
                    //out.writeObject(gbmodel);
                    //out.writeObject(rj); 
                    out.writeObject(new ObjectExamplePair("a"));
                    out.flush(); out.close(); o.close();
                } catch (Exception ex) {ex.printStackTrace();}
            }
        });
        */
        
        JMenuItem testDeterministicGoal = new JMenuItem("Test deterministic goal...");
        toolMenu.add(testDeterministicGoal);
		testDeterministicGoal.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e){
				Object[] objects = {new Float(16.25),new Float(0.0), new Float(15.5)};
				Object[] bindings = engine.deterministicGoal(
					"ip_append([97,98],[99,100],L), countList(L,N), ipObjectSpec('java.lang.Integer',Integer,[N],_), name(A,L),"+
					"assert(foofoo(Objects))", 
					"Objects",
					objects,
					"[Integer,string(A)]"
					);
				System.out.println("success="+(bindings!=null));
				System.out.println("bindings[0]="+bindings[0]);
				System.out.println("bindings[1]="+bindings[1]);
			}
		});
                
        addItemToMenu(toolMenu,"Reconsult...",new ActionListener(){
            public void actionPerformed(ActionEvent e){
                reconsultFile();
            }
        });
        
        addItemToMenu(toolMenu,"Interrupt Prolog",new ActionListener(){
            public void actionPerformed(ActionEvent e){
                engine.interrupt();
            }
        });
        
        historyMenu = new JMenu("History",true);
        mb.add(historyMenu); 
        historyMenu.addSeparator(); // to avoid Swing bug handling key events
    }
    
    class HistoryListener implements ActionListener{
        JTextComponent targetText;
        String memory;
        HistoryListener(JTextComponent t,String s){targetText=t; memory=s;}
        public void actionPerformed(ActionEvent e){
            targetText.replaceSelection(memory);
        }
    }
    
    static void addItemToMenu(JMenu menu,String item,ActionListener handler) {
        JMenuItem menuItem = new JMenuItem(item);
        menu.add(menuItem);
        menuItem.addActionListener(handler);
    }

    void sendToProlog(){
        try {
            String goal = prologInput.getText();
            prologOutput.append(goal+"\n");
            engine.sendAndFlush(goal + "\n");
            focusInput();
            //Thread.sleep(10); hack to allow the operating system to (hopefully) pipe Prolog's output...
            addToHistory();
        } catch (Exception ex) {ex.printStackTrace();}
    }

    void addToHistory(){
        JMenuItem item;
        String goal = prologInput.getText();
        if (goal.length()<2)  return; // not worthy remembering
        if (historyMenu.getItemCount() > 20) {
            historyMenu.remove(0);
            historyVector.remove(0);
        }
        historyMenu.add(item = new JMenuItem(goal));
        indexOfHistory = historyVector.size();
        historyVector.set(indexOfHistory-1, goal);
        historyVector.add("");
        //if (goal.length()>20) historyMenu.add(item = new JMenuItem(goal.substring(0,19)+"..."));
        //else historyMenu.add(item = new JMenuItem(goal.substring(0,goal.length())));
        item.addActionListener(new HistoryListener(prologInput,goal));
    }

    void reconsultFile(){
        String nome,directorio; File filetoreconsult=null;
        FileDialog d = new FileDialog(this,"Reconsult file...");
        d.show();
        nome = d.getFile(); directorio = d.getDirectory();
        if (nome!=null) {
            filetoreconsult = new File(directorio,nome);
            ;
            engine.sendAndFlush("reconsult('"+filetoreconsult.getAbsolutePath()+ "').\n");
        }
    }

    void focusInput(){
        prologInput.selectAll();
        prologInput.requestFocus();
    }
    
    // Slow startups seem to cause deadlock... 
    // perhaps we should check readiness before dumping stuff from Prolog
    public /*synchronized*/ void promptWasOutput(){
        if (debug) System.out.println("promptWasOutput()");
        scrollToBottom();
    }
    
    void scrollToBottom(){
        if (prologOutput.isShowing()) {
            prologOutput.setCaretPosition(prologOutput.getDocument().getEndPosition().getOffset()-1 /* OBOB hack */);
            try {
                // If we're in a JScrollPane, force scrolling to bottom and left
                JScrollBar scrollbarV = ((JScrollPane)((JViewport)(prologOutput.getParent())).getParent()).getVerticalScrollBar();
                scrollbarV.setValue(scrollbarV.getMaximum());
                JScrollBar scrollbarH = ((JScrollPane)((JViewport)(prologOutput.getParent())).getParent()).getHorizontalScrollBar();
                scrollbarH.setValue(scrollbarH.getMinimum());
            } catch (Exception e) {/* We're not in a JScrollPane, forget it! */};
        }
    }
    
    public /* synchronized*/ void print(String s){
        if (debug) System.out.println("print("+s+")");
        prologOutput.append(s);
        scrollToBottom();
    } 

	public static void main(String args[]) {
		System.out.println("Welcome to InterProlog 1.0b3 "+
			System.getProperty("java.version") + " ("+
			System.getProperty("java.vendor") + "), "+ 
			System.getProperty("os.name") + " "+
			System.getProperty("os.version"));
		// pass along XSB's start command; this will affect all PrologEngines:
		PrologEngine.setDefaultPrologCommand(args); 
		new TopLevelWindow();
	}
}